using System;
using System.Data.Common;
using System.Data.Odbc;
using System.ComponentModel;
using System.Data;
using System.Globalization;

namespace SqlBi.Tools.SqlBulkTool {
    /// <summary>
    /// 
    /// Implementation of DataBaseCopy that uses OdbcConnection for Navision database
    /// 
    /// </summary>
    public class NavisionDataBaseCopy : DatabaseCopy {
        public override IDbCommand GetDbCommand( string commandString ) {
            OdbcConnection conn = new OdbcConnection( SourceConnectionString );
            return new NavisionCommand( commandString, conn );
        }

        protected static string TransformOdbcName( string name ) {
            return name.Replace( ".", "_" ).Replace( ",", "_" ).Replace( " ", "_" );
        }

        /// <summary>
        /// 
        /// Returns a valid identifier for the SQL dialect implemented. 
        /// The identifier is encapsulated between delimiters and 
        /// internal characters are substituted where necessary
        /// 
        /// </summary>
        /// <param name="identifier">Identifier to normalize</param>
        /// <returns></returns>
        public override string EncapsulateSourceIdentifier( string identifier ) {
            return String.Format (CultureInfo.InvariantCulture, "\"{0}\"", TransformOdbcName (identifier));
        }
    }

    /// <summary>
    ///  
    /// Encapsulate OdbcCommand just to intercept the ExecuteReader method and return a NavisionDataReaded wrapper class
    /// 
    /// </summary>
    public class NavisionCommand : DbCommand { 
        private OdbcCommand _odbcCommand;
        public OdbcCommand OdbcCommand {
            get { return _odbcCommand; }
            set { _odbcCommand = value; }
        }

        /// <summary>
        /// Initializes a new instance of the System.Data.Odbc.OdbcCommand class.
        /// </summary>
        public NavisionCommand() {
            _odbcCommand = new OdbcCommand();
        }

        /// <summary>
        /// Initializes a new instance of the System.Data.Odbc.OdbcCommand class with the text of the query.
        /// </summary>
        /// <param name="cmdText">The text of the query</param>
        public NavisionCommand( string cmdText ) {
            _odbcCommand = new OdbcCommand( cmdText );
        }
        /// <summary>
        /// Initializes a new instance of the System.Data.Odbc.OdbcCommand class with
        /// the text of the query and an System.Data.Odbc.OdbcConnection object.
        /// </summary>
        /// <param name="cmdText">The text of the query</param>
        /// <param name="connection">An System.Data.Odbc.OdbcConnection object that represents the connection to a data source</param>
        public NavisionCommand( string cmdText, OdbcConnection connection ) {
            _odbcCommand = new OdbcCommand( cmdText, connection );
        }

        /// <summary>
        /// Initializes a new instance of the System.Data.Odbc.OdbcCommand class with
        /// the text of the query and an System.Data.Odbc.OdbcConnection object.
        /// </summary>
        /// <param name="cmdText">The text of the query</param>
        /// <param name="connection">An System.Data.Odbc.OdbcConnection object that represents the connection to a data source</param>
        /// <param name="transaction">The transaction in which the System.Data.Odbc.OdbcCommand executes</param>
        public NavisionCommand( string cmdText, OdbcConnection connection, OdbcTransaction transaction ) {
            _odbcCommand = new OdbcCommand( cmdText, connection, transaction );
        }

        /// <summary>
        /// Gets or sets the SQL statement or stored procedure to execute against the data source.
        /// </summary>
        public override string CommandText {
            get { return _odbcCommand.CommandText; }
            set { _odbcCommand.CommandText = value; }
        }


        /// <summary>
        /// Gets or sets the wait time before terminating an attempt to execute a command and generating an error.
        /// </summary>
        public override int CommandTimeout {
            get { return _odbcCommand.CommandTimeout; }
            set { _odbcCommand.CommandTimeout = value; }
        }

        /// <summary>
        /// Gets or sets a value that indicates how the System.Data.Odbc.OdbcCommand.CommandText property is interpreted.
        /// </summary>
        public override CommandType CommandType {
            get { return _odbcCommand.CommandType; }
            set { _odbcCommand.CommandType = value; }
        }

        protected override DbConnection DbConnection {
            get { return _odbcCommand.Connection; }
            set { _odbcCommand.Connection = (OdbcConnection) value; }
        }
        protected override DbParameterCollection DbParameterCollection {
            get { return _odbcCommand.Parameters; }
        }
        protected override DbTransaction DbTransaction {
            get { return _odbcCommand.Transaction; }
            set { _odbcCommand.Transaction = (OdbcTransaction)value; }
        }
        protected override DbParameter CreateDbParameter() {
            return _odbcCommand.CreateParameter();
        }
        protected override DbDataReader ExecuteDbDataReader( CommandBehavior behavior ) {
            return new NavisionDataReader( _odbcCommand.ExecuteReader( behavior ) );
        }

        /// <summary>
        ///     Gets or sets a value that indicates whether the command object should be
        ///     visible in a customized interface control.
        /// </summary>
        // [EditorBrowsable( 1 )]
        [Browsable( false )]
        [DefaultValue( true )]
        [DesignOnly( true )]
        public override bool DesignTimeVisible {
            get { return _odbcCommand.DesignTimeVisible; }
            set { _odbcCommand.DesignTimeVisible = value; }
        }

        /// <summary>
        /// Gets or sets the System.Data.Odbc.OdbcConnection used by this instance of the System.Data.Odbc.OdbcCommand.
        /// </summary>
        public new DbConnection Connection {
            get { return _odbcCommand.Connection; }
            set { _odbcCommand.Connection = (OdbcConnection)value; base.Connection = value; }
        }

        /// <summary>
        /// Gets the System.Data.Odbc.OdbcParameterCollection.
        /// </summary>
        // [DesignerSerializationVisibility( 2 )]
        public new DbParameterCollection Parameters {
            get { return _odbcCommand.Parameters; }
        }

        /// <summary>
        /// Gets or sets the System.Data.Odbc.OdbcTransaction within which the System.Data.Odbc.OdbcCommand executes.
        /// </summary>
        [DesignerSerializationVisibility( 0 )]
        [Browsable( false )]
        public new DbTransaction Transaction {
            get { return _odbcCommand.Transaction; }
            set { _odbcCommand.Transaction = (OdbcTransaction)value; base.Transaction = value; }
        }

        /// <summary>
        /// Gets or sets a value that specifies how the Update method should apply command results to the DataRow.
        /// </summary>
        public override UpdateRowSource UpdatedRowSource {
            get { return _odbcCommand.UpdatedRowSource; }
            set { _odbcCommand.UpdatedRowSource = value; }
        }

        /// <summary>
        /// Tries to cancel the execution of an System.Data.Odbc.OdbcCommand.
        /// </summary>
        public override void Cancel() {
            _odbcCommand.Cancel();
        }

        //
        // Summary:
        //     Creates a new instance of an System.Data.Odbc.OdbcParameter object.
        //
        // Returns:
        //     An System.Data.Odbc.OdbcParameter object.
        public new DbParameter CreateParameter() {
            return _odbcCommand.CreateParameter();
        }

        /// <summary>
        /// Executes an SQL statement against the System.Data.Odbc.OdbcCommand.Connection and returns the number of rows affected.
        /// </summary>
        /// <returns>
        /// For UPDATE, INSERT, and DELETE statements, the return value is the number
        /// of rows affected by the command. For all other types of statements, the return
        /// value is -1.
        /// </returns>
        public override int ExecuteNonQuery() {
            return _odbcCommand.ExecuteNonQuery();
        }

        /// <summary>
        ///     Sends the System.Data.Odbc.OdbcCommand.CommandText to the System.Data.Odbc.OdbcCommand.Connection
        ///     and builds an System.Data.Odbc.OdbcDataReader.
        /// </summary>
        /// <returns>
        ///     An System.Data.Odbc.OdbcDataReader object.
        /// </returns>
        public new NavisionDataReader ExecuteReader() {
            // TODO : Fare wrapper qui
            return new NavisionDataReader( _odbcCommand.ExecuteReader() );
        }

        /// <summary>
        ///     Sends the System.Data.Odbc.OdbcCommand.CommandText to the System.Data.Odbc.OdbcCommand.Connection,
        ///     and builds an System.Data.Odbc.OdbcDataReader using one of the CommandBehavior
        ///     values.
        /// </summary>
        /// <param name="behavior">One of the System.Data.CommandBehavior values.</param>
        /// <returns>
        ///     An System.Data.Odbc.OdbcDataReader object.
        /// </returns>
        public new NavisionDataReader ExecuteReader( CommandBehavior behavior ) {
            // TODO : Fare wrapper qui
            return new NavisionDataReader( _odbcCommand.ExecuteReader( behavior ) );
        }

        /// <summary>
        ///     Executes the query, and returns the first column of the first row in the
        ///     result set returned by the query. Additional columns or rows are ignored.
        /// </summary>
        /// <returns>
        ///     The first column of the first row in the result set, or a null reference
        ///     if the result set is empty.
        /// </returns>
        public override object ExecuteScalar() {
            return _odbcCommand.ExecuteScalar();
        }

        /// <summary>
        /// Creates a prepared or compiled version of the command at the data source.
        /// </summary>
        public override void Prepare() {
            _odbcCommand.Prepare();
        }

        /// <summary>
        /// Resets the System.Data.Odbc.OdbcCommand.CommandTimeout property to the default value.
        /// </summary>
        public void ResetCommandTimeout() {
            _odbcCommand.ResetCommandTimeout();
        } 
    }

    /// <summary>
    /// Encapsulate OdbcDataReader to filter access to special Navision fields (ie DateTime)
    /// </summary>
    public class NavisionDataReader : DbDataReader {
        private static DateTime MinDateTime = new DateTime( 1900, 1, 1 );

        public override bool HasRows {
            get { return wrappedReader.HasRows; }
        }

        public override System.Collections.IEnumerator GetEnumerator() {
            return wrappedReader.GetEnumerator();
        }

        #region IDataReader Members

        public override void Close() {
            wrappedReader.Close();
        }

        public override int Depth {
            get { return wrappedReader.Depth; }
        }

        public override DataTable GetSchemaTable() {
            return wrappedReader.GetSchemaTable();
        }

        public override bool IsClosed {
            get { return wrappedReader.IsClosed; }
        }

        public override bool NextResult() {
            return wrappedReader.NextResult();
        }

        public override bool Read() {
            return wrappedReader.Read();
        }

        public override int RecordsAffected {
            get { return wrappedReader.RecordsAffected; }
        }

        #endregion

        #region IDataRecord Members

        public override int FieldCount {
            get { return wrappedReader.FieldCount; }
        }

        public override bool GetBoolean( int i ) {
            return wrappedReader.GetBoolean( i );
        }

        public override byte GetByte( int i ) {
            return wrappedReader.GetByte( i );
        }

        public override long GetBytes( int i, long fieldOffset, byte[] buffer, int bufferoffset, int length ) {
            return wrappedReader.GetBytes( i, fieldOffset, buffer, bufferoffset, length );
        }

        public override char GetChar( int i ) {
            return wrappedReader.GetChar( i );
        }

        public override long GetChars( int i, long fieldoffset, char[] buffer, int bufferoffset, int length ) {
            return wrappedReader.GetChars( i, fieldoffset, buffer, bufferoffset, length );
        }

        public new DbDataReader GetData( int i ) {
            return wrappedReader.GetData( i );
        }

        public override string GetDataTypeName( int i ) {
            return wrappedReader.GetDataTypeName( i );
        }

        public override DateTime GetDateTime( int i ) {
            return wrappedReader.GetDateTime( i );
        }

        public override decimal GetDecimal( int i ) {
            return wrappedReader.GetDecimal( i );
        }

        public override double GetDouble( int i ) {
            return wrappedReader.GetDouble( i );
        }

        public override Type GetFieldType( int i ) {
            return wrappedReader.GetFieldType( i );
        }

        public override float GetFloat( int i ) {
            return wrappedReader.GetFloat( i );
        }

        public override Guid GetGuid( int i ) {
            return wrappedReader.GetGuid( i );
        }

        public override short GetInt16( int i ) {
            return wrappedReader.GetInt16( i );
        }

        public override int GetInt32( int i ) {
            return wrappedReader.GetInt32( i );
        }

        public override long GetInt64( int i ) {
            return wrappedReader.GetInt64( i );
        }

        public override string GetName( int i ) {
            return wrappedReader.GetName( i );
        }

        public override int GetOrdinal( string name ) {
            return wrappedReader.GetOrdinal( name );
        }

        public override string GetString( int i ) {
            return wrappedReader.GetString( i );
        }

        public override object GetValue( int i ) {
            object o = wrappedReader.GetValue( i );
            if (o is DateTime) {
                if (((DateTime) o) < MinDateTime) {
                    o = null;
                }
            }
            return o;
            //return wrappedReader.GetValue( i );
        }

        public override int GetValues( object[] values ) {
            return wrappedReader.GetValues( values );
        }

        public override bool IsDBNull( int i ) {
            return wrappedReader.IsDBNull( i );
        }

        public override object this[string name] {
            get { return wrappedReader[name]; }
        }

        public override object this[int i] {
            get { return wrappedReader[i]; }
        }

        #endregion

        private OdbcDataReader wrappedReader;

        public NavisionDataReader( OdbcDataReader odbcDataReader ) {
            wrappedReader = odbcDataReader;
        }
        protected override void Dispose( bool disposing ) {
            if (disposing) {
                wrappedReader.Dispose();
            }
            base.Dispose( disposing );
        }

        ~NavisionDataReader() {
            this.Dispose( false );
        }

    }

}
