using System;
using System.Collections.Generic;
using System.Text;
using System.Data;
using System.Data.Common;
using System.Data.SqlClient;
using System.Data.OleDb;
using System.Threading;
using System.Globalization;

namespace SqlBi.Tools.SqlBulkTool {
    /// <summary>
    /// 
    /// This class contains information about one partition of a table and handles the necessary operation
    /// to do partition switching and database loading.
    /// 
    /// The correct and fast operation to perform is:
    /// 
    /// - Switch out the partition to a temporary table 
    /// - Truncate the temporary table
    /// - Load data from the source to the temporary table
    /// - Add a CHECK contraint to ensure that the data in the temporary table is correct
    /// - Switch in the temporary table into the partition
    /// - Drop the temporary table.
    /// 
    /// and is handled by three methods:
    /// 
    /// - SwitchOut : creates the temporary table and switches partition out
    /// - Load      : loads data from the source
    /// - Switch In : adds the contraints, switches in the partition and then drop it
    /// 
    /// </summary>
    public class PartitionCopy {
        private TableCopy _Table;            // The table that contains us
        private string _PartitionId;         // The partition ID
        private IComparable _MinValue;       // The boundary Value left
        private IComparable _MaxValue;       // The boundary Value right
        private bool _BoundaryValueOnRight;  // When BoundaryValueOnRight then partition contains data with key < Value, otherwise
                                             // Partition contains data with key >= value
        private bool _NeedsRebuild;          // Do we need to rebuild this partition?

        public bool NeedsRebuild {
            get { return _NeedsRebuild; }
            set { _NeedsRebuild = value; }
        }
        public IComparable MinValue {
            get { return _MinValue; }
            set { _MinValue = value; }
        }
        public IComparable MaxValue {
            get { return _MaxValue; }
            set { _MaxValue = value; }
        }
        public string PartitionId {
            get { return _PartitionId; }
            set { _PartitionId = value; }
        }
        public bool BoundaryValueOnRight {
            get { return _BoundaryValueOnRight; }
            set { _BoundaryValueOnRight = value; }
        }
        public TableCopy Table {
            get { return _Table; }
            set { _Table = value; }
        }

        /// <summary>
        /// 
        /// Standard constructor for Partition, it receives the table that contains the partition as a parameter
        /// 
        /// </summary>
        /// <param name="pOwner"></param>
        public PartitionCopy (TableCopy table) {
            Table = table;
        }

        /// <summary>
        /// 
        /// Process is a thread safe implementation of the copy operation for one partition.
        /// 
        /// It locks Table.MetaData performing each metadata update, using this tecnique we can perform in parallel 
        /// the LoadPartition function that is very time consuming still be assured that switch in and out will be 
        /// carried on sequentially.
        /// 
        /// At the end of the operation we Signal the Processed Event to let the caller know that the operation has
        /// been carried on correctly. Beware that if Procesed is not set correctly the entire process will be in 
        /// a deadlock situation.
        /// 
        /// Please note that the process wait for the database to start and, at its end, communicates the database
        /// its termination, in this way the database can decide to limit the number of parallel threads that
        /// runs in a period of time. See DataBase to have more information about it.
        /// 
        /// </summary>
        public void Process () {
            if (Table.TruncateBeforeLoad) {
                SwitchOut ();         // Metadata update is in critical section for the table
                Load (true);          // Load operation is outside critical section
                SwitchIn ();          // Metadata update is in critical section for the table
            } else {
                Load (false);         // Load operation if no truncate is needed
            }
        }

        /// <summary>
        /// 
        /// Creates a new temporary table, then switch out a partition from the destination table
        /// and truncates the temporary table so we will be able to add data from the source table
        /// into the temporary.
        /// 
        /// </summary>
        private void SwitchOut () {
            string DropPartitionTable = String.Format (CultureInfo.InvariantCulture, @"IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID('{0}') AND type in (N'U')) DROP TABLE {0}", GetTempTableName ());
            string CreateNewPartitionTable = String.Format (CultureInfo.InvariantCulture, "Select * Into [{0}] From [{1}].[{2}] Where 1 = 0 ", GetTempTableName (), Table.DestTableSchema, Table.DestTableName);
            string TruncateTable = String.Format (CultureInfo.InvariantCulture, "Truncate Table {0}", GetTempTableName ());
            string SwitchOutPartition = String.Format (CultureInfo.InvariantCulture, "Alter Table [{0}].[{1}] Switch Partition {2} to [{3}]", Table.DestTableSchema, Table.DestTableName, PartitionId, GetTempTableName ());

            Table.Database.RunCriticalCommands (Table, DropPartitionTable, CreateNewPartitionTable, SwitchOutPartition, TruncateTable);
        }

        /// <summary>
        /// 
        /// Loads data from the source table to the temporary destination table, it builds a SELECT statement
        /// that contains only the useful fields.
        /// 
        /// </summary>
        private void Load (bool intoPartition) {
            using (SqlConnection DestConn = Table.Database.GetDestConn ()) {
                IDbCommand SourceCommand = Table.Database.GetSourceCommand (Table, GetSourceCondition ());
                //
                //  Now we can open a datareader on the source connection and invoke BulkCopy to transfer data to the
                //  temporary table
                //
                DestConn.Open ();
                SourceCommand.Connection.Open ();
                try {
                    using (IDataReader SourceReader = SourceCommand.ExecuteReader ()) {
                        using (SqlBulkCopy bulkCopy = new SqlBulkCopy (DestConn)) {
                            if (intoPartition) {
                                bulkCopy.DestinationTableName = GetTempTableName ();
                            } else {
                                bulkCopy.DestinationTableName = String.Format (CultureInfo.InvariantCulture, "[{0}].[{1}]", Table.DestTableSchema, Table.DestTableName);
                            }
                            bulkCopy.BulkCopyTimeout = 3600;
                            bulkCopy.BatchSize = 5000;

                            bulkCopy.WriteToServer (SourceReader);

                            bulkCopy.Close ();
                            SourceReader.Close ();
                        }
                    }
                } finally {
                    SourceCommand.Connection.Close ();
                }
            }
        }

        /// <summary>
        /// 
        /// Adds the necessary CHECK constraint to the temporary table, then switches it in as a
        /// partition of the destination table. At the end it DROPs the temporary table that is
        /// no longer usefule
        /// 
        /// </summary>
        private void SwitchIn () {
            string AddConstraint = String.Format (CultureInfo.InvariantCulture, "Alter Table [{0}] WITH CHECK Add Constraint [{0}_CK] Check ({1})", GetTempTableName (), GetDestinationCondition ());
            string SwitchInPartition = String.Format (CultureInfo.InvariantCulture, "Alter Table [{0}] Switch To [{1}].[{2}] Partition {3}", GetTempTableName (), Table.DestTableSchema, Table.DestTableName, PartitionId);
            string DropTable = String.Format (CultureInfo.InvariantCulture, "DROP TABLE [{0}]", GetTempTableName ());

            Table.Database.RunCriticalCommands (Table, AddConstraint, SwitchInPartition, DropTable);
        }

        /// <summary>
        /// 
        /// Returns the name of the temporary table that is used to load the data before the switch 
        /// 
        /// </summary>
        /// <returns></returns>
        private string GetTempTableName () {
            return String.Format (CultureInfo.InvariantCulture, "{0}_Partition_{1}", Table.DestTableName, PartitionId);
        }

        /// <summary>
        /// 
        /// Checks if this partition contains any data between UpdateMin and UpdateMax, it is used to determine
        /// if the partition needs a rebuild.
        /// 
        /// </summary>
        /// <param name="UpdateMinValue"></param>
        /// <param name="UpdateMaxValue"></param>
        /// <returns></returns>
        public bool ContainsDataBetween (IComparable updateMinValue, IComparable updateMaxValue) {
            if ((MinValue != null) && (MaxValue != null)) {
                return (MaxValue.CompareTo (updateMinValue) > 0) && (MinValue.CompareTo (updateMaxValue) < 0);
            } else {
                if (MinValue != null) {
                    return (MinValue.CompareTo (updateMaxValue) < 0);
                } else {
                    return (MaxValue.CompareTo (updateMinValue) > 0);
                }
            }
        }

        /// <summary>
        /// 
        /// Returns the condition of the data that are contained in the partition, it is used both to filter
        /// source table and to impose a check condition to the destinatino table prior to switch in into
        /// the partitioned table and so is called by both GetSourceCondition and GetDestinationCondition
        /// 
        /// </summary>
        /// <returns></returns>
        private string GetCondition (string ColName) {
            if (BoundaryValueOnRight) {
                if ((MinValue != null) && (MaxValue != null)) {
                    return String.Format (CultureInfo.InvariantCulture, "{0} >= {1} And {0} < {2}", ColName, MinValue, MaxValue);
                } else {
                    if (MinValue != null) {
                        return String.Format (CultureInfo.InvariantCulture, "{0} >= {1}", ColName, MinValue);
                    } else {
                        return String.Format (CultureInfo.InvariantCulture, "{0} < {1}", ColName, MaxValue);
                    }
                }
            } else {
                //
                //  TODO: code to review
                //
                if ((MinValue != null) && (MaxValue != null)) {
                    return String.Format (CultureInfo.InvariantCulture, "{0} >= {1} And {0} < {2}", ColName, MinValue, MaxValue);
                } else {
                    if (MinValue != null) {
                        return String.Format (CultureInfo.InvariantCulture, "{0} >= {1}", ColName, MinValue);
                    } else {
                        return String.Format (CultureInfo.InvariantCulture, "{0} < {1}", ColName, MaxValue);
                    }
                }
            }
        }

        /// <summary>
        /// 
        /// Returns a valid condition to use against the source database
        /// 
        /// </summary>
        /// <returns></returns>
        private string GetSourceCondition () {
            return this.GetCondition (Table.Database.EncapsulateSourceIdentifier (Table.PartitionColumnName));
        }

        /// <summary>
        /// 
        /// Returns a valid condition to use against the destination database (always SQL Server 2005)
        /// 
        /// </summary>
        /// <returns></returns>
        private string GetDestinationCondition () {
            return this.GetCondition (Table.Database.EncapsulateDestinationIdentifier (Table.PartitionColumnName));
        }
    }
}
